package main

import (
	"fmt"
	"io"
	"os"
	"path/filepath"
	"regexp"

	"github.com/avenga/couper/errors"
)

var couperErrors []string

func main() {
	println("generating error types:\n")

	// hold list of possible base kinds which can be skipped for type definitions
	b, err := os.ReadFile(filepath.Join("errors", "couper.go"))
	must(err)
	result := regexp.MustCompile(`\t(\w+)\s+= &Error.+\n`).FindAllStringSubmatch(string(b), -1)
	for _, r := range result {
		couperErrors = append(couperErrors, r[1])
	}

	errTypes := map[string]int{}
	sortedErrTypes := make([]string, len(errors.Definitions))
	for idx, errImpl := range errors.Definitions {
		kinds := errImpl.Kinds()
		if len(kinds) == 0 {
			must(fmt.Errorf("error kind must be defined"))
		}

		kind := kinds[0]
		if _, exist := errTypes[kind]; exist {
			must(fmt.Errorf("error kind already defined: %s", kind))
		}
		errTypes[kind] = idx
		sortedErrTypes[idx] = kind
	}

	for _, kind := range sortedErrTypes {
		println("\t", kind, "-->", errors.SnakeToCamel(kind))
	}

	generated, err := os.Create(filepath.Join("errors", "types_generated.go"))
	must(err)
	defer generated.Close()

	_, err = io.WriteString(generated, fmt.Sprintf(`// Code generated by go generate; DO NOT EDIT.

package errors

var (
`))
	must(err)

	for idx, kind := range sortedErrTypes {
		if isDefined(kind) {
			continue
		}
		_, err = io.WriteString(generated, fmt.Sprintf("\t%s = Definitions[%d]\n", errors.SnakeToCamel(kind), idx))
		must(err)
	}

	_, err = io.WriteString(generated, fmt.Sprintf(`)

// typeDefinitions holds all related error definitions which are
// catchable with an error_handler definition.
type typeDefinitions map[string]*Error

// types holds all implemented ones. The name must match the structs
// snake-name for fallback purposes. See TypeToSnake usage and reference.
var types = typeDefinitions{
`))

	must(err)

	for _, kind := range sortedErrTypes {
		_, err = io.WriteString(generated, fmt.Sprintf("\t%q: %s,\n", kind, errors.SnakeToCamel(kind)))
		must(err)
	}

	_, err = io.WriteString(generated, fmt.Sprintf(`}

// IsKnown tells the configuration callee if Couper
// has a defined error type with the given name.
func IsKnown(errorType string) bool {
	_, known := types[errorType]
	return known
}

`))

	must(err)

}

// isDefined checks if the given typeName is defined and must be skipped for declaration with true result.
func isDefined(typeName string) bool {
	for _, t := range couperErrors {
		if errors.SnakeToCamel(typeName) == t {
			return true
		}
	}
	return false
}

func must(err error) {
	if err != nil {
		println(err.Error())
		os.Exit(1)
	}
}
